﻿using HslCommunication;
using HslCommunication.BasicFramework;
using HslCommunication.Core;
using HslCommunication.Core.IMessage;
using HslCommunication.Core.Net;
using ProjectManagementSystem.Common.Logger;
using ProjectManagementSystem.Device.Core;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace ProjectManagementSystem.Common.Profinet
{
    public class HsjmPlcNet : NetworkDeviceBase<HsjmPlcMessage, RegularByteTransform>
    {
        private readonly int total;

        private byte[] localData;
        private byte[] readCommand;

        public HsjmPlcNet(string ipAddress, int port, int total)
        {
            IpAddress = ipAddress;
            Port = port;

            this.total = total;
            localData = new byte[total];

            readCommand = new byte[total + 1];
            readCommand[0] = 0x01;
        }

        public override OperateResult<byte[]> Read(string address, ushort length)
        {
            // 获取指令
            if(!int.TryParse(address.Substring(1), out int idx))
            {
                return OperateResult.CreateFailedResult<byte[]>(new OperateResult($"解析地址失败 {address}"));
            }

            //用于检查写入时，直接返回本地数据
            if (address.StartsWith("A"))
            {
                return OperateResult.CreateSuccessResult(localData.Skip(idx).ToArray());
            }

            // 核心交互
            var read = ReadFromCoreServer(readCommand);
            if (!read.IsSuccess) 
            {
                if (read.Content == null || read.Content.Length == 0)
                {
                    read.Content = new byte[length];
                }
                return read;
            }

            // 数据解析
            return OperateResult.CreateSuccessResult(SoftBasic.BytesArrayRemoveBegin(read.Content, 3 + idx));
        }

        public override OperateResult<bool[]> ReadBool(string address, ushort length)
        {
            // 核心交互
            var read = ReadFromCoreServer(readCommand);
            if (!read.IsSuccess) return OperateResult.CreateFailedResult<bool[]>(read);

            var extract = OperateResult.CreateSuccessResult(SoftBasic.BytesArrayRemoveBegin(read.Content, 3));
            if (!extract.IsSuccess) return OperateResult.CreateFailedResult<bool[]>(extract);

            // 数据解析
            return OperateResult.CreateSuccessResult(extract.Content.Select(m => m == 0x01).Take(length).ToArray());
        }


        public override OperateResult Write(string address, byte[] value)
        {
            // 解析指令
            if (!int.TryParse(address.Substring(1), out int idx))
            {
                return OperateResult.CreateFailedResult<byte[]>(new OperateResult($"解析地址失败 {address}"));
            }
            Array.Copy(value, 0, localData, idx, value.Length);

            var writeCommand = new byte[total + 1];
            writeCommand[0] = 0x02;
            Array.Copy(localData, 0, writeCommand, 1, localData.Length);

            // 核心交互
            OperateResult<byte[]> read = ReadFromCoreServer(writeCommand);
            if (!read.IsSuccess) return read;

            // 成功
            return OperateResult.CreateSuccessResult();
        }
    }

    public class HsjmPlcMessage : INetMessage
    {
        public int ProtocolHeadBytesLength => 3;


        public byte[] HeadBytes { get; set; }
        public byte[] ContentBytes { get; set; }
        public byte[] SendBytes { get; set; }

        public bool CheckHeadBytesLegal(byte[] token)
        {
            if (HeadBytes != null)
            {
                if (HeadBytes[0] == 0x05 && HeadBytes[1] == 0x06) { return true; }
            }
            return false;
        }

        public int GetContentLengthByHeadBytes()
        {
            return HeadBytes[2] - ProtocolHeadBytesLength;
        }

        public int GetHeadBytesIdentity()
        {
            return 0;
        }
    }
}
