﻿using Prism.Commands;
using Prism.Events;
using Prism.Mvvm;
using Prism.Regions;
using ProjectManagementSystem.Common.Extenions;
using ProjectManagementSystem.Common.Function;
using ProjectManagementSystem.Common.Logger;
using ProjectManagementSystem.Common.Service;
using ProjectManagementSystemView.Infrastructure.Events;
using ProjectManagementSystemView.Infrastructure.Models;
using PropertyChanged;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;

namespace ProjectManagementSystemView.ViewModels
{
    [AddINotifyPropertyChangedInterface]
    public class MainViewModel
    {
        private readonly IRegionManager regionManager;
        private readonly IEventAggregator eventAggregator;
        private IRegionNavigationJournal journal;
        
        private string tabPage;

        public string Title { get; set; } = "PMS";
        public string ConnectionIcon { get; set; } = "LanDisconnect";
        public DateTime NowTime { get; set; }

        public ObservableCollection<MenuBar> MenuBars { get; set; }

        public DelegateCommand LoadedCommand { get; private set; }
        public DelegateCommand<MenuBar> NavigateCommand { get; private set; }
        public DelegateCommand GoBackCommand { get; private set; }
        public DelegateCommand GoForwardCommand { get; private set; }
        public DelegateCommand PushMainWindow2Top { get; private set; }
        public DelegateCommand ShutdownApp { get; private set; }

        public MainViewModel(IRegionManager regionManager, IEventAggregator eventAggregator)
        {
            this.regionManager = regionManager;
            this.eventAggregator = eventAggregator;


            MenuBars = new ObservableCollection<MenuBar>();

            LoadedCommand = new DelegateCommand(Init);
            NavigateCommand = new DelegateCommand<MenuBar>(Navigate);
            GoBackCommand = new DelegateCommand(() =>
            {
                if (journal != null && journal.CanGoBack)
                    journal.GoBack();
            });
            GoForwardCommand = new DelegateCommand(() =>
            {
                if (journal != null && journal.CanGoForward)
                    journal.GoForward();
            });
            PushMainWindow2Top = new DelegateCommand(() =>
            {
                if (!string.IsNullOrEmpty(tabPage))
                {
                    if(Application.Current.MainWindow.WindowState == WindowState.Minimized)
                    {
                        Application.Current.MainWindow.WindowState = WindowState.Normal;
                    }
                    
                    Application.Current.MainWindow.Visibility = Visibility.Visible;
                    Application.Current.MainWindow.Topmost = true;
                    Application.Current.MainWindow.Topmost = false;
                }
            });
            ShutdownApp = new DelegateCommand(Shutwown);

            eventAggregator.GetEvent<UpdateViewEvent>().Subscribe(() =>
            {
                ConnectionIcon = Crms.PmsApi.Status ? "LanConnect" : "LanDisconnect";

                NowTime = DateTime.Now;
            });
        }

        private void Init()
        {
            var version = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version;
            var dateTime = System.IO.File.GetLastWriteTime(this.GetType().Assembly.Location).ToString("yyyyMMddHHmmss");
            Title = $"PMS{version} ({dateTime})";

            CLog.Instance.SystemLog.WriteInfo($"开始启动，版本：{Title}");

            var systemEnteranceName = AppSetting.TryGetValue<string>("SystemEnteranceName");
            if (!string.IsNullOrEmpty(systemEnteranceName))
            {
                var assemblyName = this.GetType().Assembly.GetName().Name;
                string fullClassName = $"{assemblyName}.Entrance.{systemEnteranceName}";
                var systemEnterance = InstanceConstructor.GetInstance<object>(assemblyName, fullClassName);
                if (systemEnterance == null)
                {
                    MessageBox.Show($"创建“{systemEnteranceName}”失败，程序退出！！！");

                    Shutwown();
                    return;
                }
            }

            CreateMenuBar();
            if (MenuBars.Count > 0)
            {
                Navigate(MenuBars.First());
                Application.Current.MainWindow.WindowState = WindowState.Normal;
            }
            else
            {
                Application.Current.MainWindow.Visibility = Visibility.Collapsed;
            }

            Task.Factory.StartNew(() =>
            {
                CLog.Instance.SystemLog.WriteDebug($"UpdateViewEvent已启动");
                while (true)
                {
                    try
                    {
                        eventAggregator.GetEvent<UpdateViewEvent>().Publish();
                    }
                    catch (Exception ex)
                    {
                        CLog.Instance.SystemLog.WriteException("UpdateViewEvent", ex);
                        Thread.Sleep(5000);
                    }
                    Thread.Sleep(1000);
                }

            });

            CLog.Instance.SystemLog.WriteInfo($"启动完毕");
        }

        private void Navigate(MenuBar obj)
        {
            if (obj == null || string.IsNullOrWhiteSpace(obj.NameSpace)) return;

            if (!regionManager.Regions.ContainsRegionWithName("MainViewRegion")) return;

            regionManager.Regions["MainViewRegion"]?.RequestNavigate(obj.NameSpace, back =>
            {
                journal = back.Context.NavigationService.Journal;
            });
        }

        void CreateMenuBar()
        {
            tabPage = AppSetting.TryGetValue<string>("TabPage");
            string[] tabPageArray = tabPage.ToValueArray<string>();
            for (int i = 0; i < tabPageArray.Length; i++)
            {
                string pagInfoStr = tabPageArray[i];
                string[] pageInfos = pagInfoStr.Split(new char[] { ':' });
                string page = pageInfos[0];
                string pageName = pageInfos[1];

                MenuBars.Add(new MenuBar() { Title = pageName, NameSpace = page });
            }
        }

        private void Shutwown()
        {
            Application.Current.Shutdown();

            Process[] process = Process.GetProcesses();
            Process currentProcess = Process.GetCurrentProcess();
            for (int i = 0; i < process.Count(); i++)
            {
                if (process[i].ProcessName == currentProcess.ProcessName)
                {
                    process[i].Kill();
                }
            }
        }
    }
}
